#include <PalmOSARM.h>
#include <EntryNums/BootEntryNums.h>
#include "TwDevAccess.h"
#include "TwDevAccessRsc.h"

#define kPalmOS527TVersion	sysMakeROMVersion(5, 2, 7, sysROMStageDevelopment, 0)

AppGlobalsType *GetGlobalsPtr(void)
{
	AppGlobalsType *gP;
	Err err;

	err = FtrGet(appFtrCreator, appFtrNumGlobalsPtr, (UInt32 *)&gP);
	return err ? NULL : gP;
}

static void *GetObjectPtr(FormPtr formP, UInt16 objectID)
{
	return FrmGetObjectPtr(formP, FrmGetObjectIndex(formP, objectID));
}

static void InstallPatches(void)
{
	AppGlobalsType *gP = GetGlobalsPtr();
	MemHandle resH, dbH;
	UInt32 sysRefNum;
	DmOpenRef dbP;
	Err err;

	err = SysFindModule(sysFileTSystem, sysFileCSystem, 0, 0, &sysRefNum);
	if (err)
		return;

	err = RALPatchEntry(sysRefNum, entryNumDmFindDatabaseWithTypeCreator,
			DmFindDatabaseWithTypeCreatorPatch,
			(void **)&gP->oldDmFindDatabaseWithTypeCreatorP);
	if (err)
		return;

	SysCurAppDatabase(&dbH);
	DmDatabaseProtect(dbH, true);

	dbP = DmOpenDatabaseByTypeCreator(sysFileTSystem, sysFileCDAL,
			dmModeReadOnly);
	resH = DmGet1Resource(sysResTBootCode, sysResIDBootHAL);
	gP->dalCodeStartP = MemHandleLock(resH);
	gP->dalCodeEndP = (UInt8 *)gP->dalCodeStartP + MemHandleSize(resH);
	MemHandleUnlock(resH);
	DmReleaseResource(resH);
	DmCloseDatabase(dbP);
}

static AppGlobalsType *AllocGlobals(void)
{
	AppGlobalsType *gP;

	gP = MemPtrNew(sizeof(*gP));
	MemSet(gP, sizeof(*gP), 0);
	MemPtrSetOwner(gP, 0);
	FtrSet(appFtrCreator, appFtrNumGlobalsPtr, (UInt32)gP);

	return gP;
}

static void EnableAccess(void)
{
	AppGlobalsType *gP = GetGlobalsPtr();
	if (gP == NULL) {
		gP = AllocGlobals();
		InstallPatches();
	}
	gP->accessEnabled = true;
}

static void DisableAccess(void)
{
	AppGlobalsType *gP = GetGlobalsPtr();
	if (gP != NULL) {
		gP->accessEnabled = false;
		FrmAlert(AccessDisabledAlert);
	}
}

static void MainFormInit(FormPtr formP)
{
	AppGlobalsType *gP;
	MemHandle memH;
	Char *stringP;
	void *dataP;
	UInt16 size;
	Err err;

	gP = GetGlobalsPtr();
	CtlSetValue(GetObjectPtr(formP, MainFormAccessEnabledCheckbox),
			(gP != NULL) && gP->accessEnabled);
	err = SysGetROMToken(sysROMTokenSnum, &dataP, &size);
	if (err)
		return;

	memH = MemHandleNew(size + 1);
	stringP = MemHandleLock(memH);
	MemMove(stringP, dataP, size);
	stringP[size] = '\0';
	MemHandleUnlock(memH);
	FldSetTextHandle(GetObjectPtr(formP, MainFormSerialNumberField),
			memH);
}

static Boolean MainFormDoCommand(UInt16 command)
{
	Boolean handled = false;
	TwGfxInfoType info;
	TwGfxHandle gfxH;
	FormPtr formP;
	Err err;

	switch (command) {
	case MenuOptionsAbout:
		MenuEraseStatus(NULL);
		formP = FrmInitForm(AboutForm);
		FrmDoDialog(formP);
		FrmDeleteForm(formP);
		handled = true;
		break;

	case MenuOptionsTestAccess:
		MenuEraseStatus(NULL);
		info.reserved[0] = sizeof(info);
		err = TwGfxOpen(&gfxH, &info);
		if (err == errNone) {
			FrmAlert(AccessAllowedAlert);
			TwGfxClose(gfxH);
		}
		handled = true;
		break;
	}

	return handled;
}

static Boolean MainFormHandleEvent(EventType *eventP)
{
	Boolean handled = false;
	FormPtr formP;
	Int16 value;

	switch (eventP->eType) {
	case ctlSelectEvent:
		switch (eventP->data.ctlSelect.controlID) {
		case MainFormAccessEnabledCheckbox:
			formP = FrmGetActiveForm();
			value = FrmGetControlValue(formP, FrmGetObjectIndex(formP, MainFormAccessEnabledCheckbox));
			if (value)
				EnableAccess();
			else
				DisableAccess();
			break;
		}
		break;

	case menuEvent:
		handled = MainFormDoCommand(eventP->data.menu.itemID);
		break;

	case frmOpenEvent:
		formP = FrmGetActiveForm();
		MainFormInit(formP);
		FrmDrawForm(formP);
		handled = true;
		break;
	}

	return handled;
}

static Boolean AppHandleEvent(EventType *eventP)
{
	FormPtr formP;
	UInt16 formID;

	if (eventP->eType == frmLoadEvent) {
		formID = eventP->data.frmLoad.formID;
		formP = FrmInitForm(formID);
		FrmSetActiveForm(formP);

		switch (formID) {
		case MainForm:
			FrmSetEventHandler(formP, MainFormHandleEvent);
			break;
		}
		return true;
	}
	return false;
}

static Err AppStart(void)
{
	return errNone;
}

static void EventLoop(void)
{
	EventType event;
	Err err;

	do {
		SysEventGet(&event, 0);

		if (!SysHandleEvent(&event)) {
			if (!MenuHandleEvent(NULL, &event, &err)) {
				if (!AppHandleEvent(&event))
					FrmDispatchEvent(&event);
			}
		}
	} while (event.eType != appStopEvent);
}

static void AppStop(void)
{
	FrmCloseAllForms();
}

static Err RomVersionCompatible(UInt32 requiredVersion, UInt16 launchFlags)
{
	UInt32 romVersion;

	FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);
	if (romVersion < requiredVersion) {
		if ((launchFlags & (sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp)) ==
				(sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp)) {
			FrmAlert(SystemIncompatibleAlert);
		}
		return sysErrRomIncompatible;
	}

	return errNone;
}

UInt32 PilotMain(UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags)
{
	Err err;

	err = RomVersionCompatible(kPalmOS527TVersion, launchFlags);
	if (err)
		return err;

	switch (cmd) {
	case sysAppLaunchCmdNormalLaunch:
		err = AppStart();
		if (err == errNone) {
			FrmGotoForm(MainForm);
			EventLoop();
			AppStop();
		}
		break;
	}
	return errNone;
}
