PRCFLAGS += -dbFlagExecutable

###############################################################################
# Defaults

ifeq ($(OBJ_DIR), )
	OBJ_DIR := _obj
endif

ifeq ($(RSC_OBJ_DIR), )
	RSC_OBJ_DIR = _rsc_obj
endif

# Locale (defaults to english)
ifeq ($(LOCALE), )
	LOCALE := enUS
endif

ifeq ($(TARGET_PLATFORM), )
	TARGET_PLATFORM := device_psi
else
ifeq ($(TARGET_PLATFORM), device)
	TARGET_PLATFORM := device_psi
endif
endif

###############################################################################
# PRC and Database Names
#
ifeq ($(PRC_NAME), )
	PRC_NAME = NativeApp.prc
endif

ifeq ($(DATABASE_NAME), )
	DATABASE_NAME = NativeApp
endif


###############################################################################
# Compile and Link Settings

TARGET_FORMAT = PalmOS5

ifeq ($(TARGET_PLATFORM), device_psi)
	CC = pacc
	AS = paasm
	AR = palib
	CFLAGS += -nologo -wd14 -Xaddr=SHL -Xgnum -Xc+=c99 \
			-D__arm__ -D"__PALMOS__=1" -DCPU_TYPE=3
	ASFLAGS +=
else
ifeq ($(TARGET_PLATFORM), device_gnu)
	CC = arm-palmos-gcc
	AS = arm-palmos-as
	AR = arm-palmos-ar
	CFLAGS += -Wmissing-prototypes -Wmissing-declarations \
			-Wno-multichar -Wno-unused-function \
			-nostdinc -fno-builtin -fshort-enums \
			-fomit-frame-pointer -fpic -ffixed-r9 \
			-mno-apcs-frame -msingle-pic-base \
			-mstructure-size-boundary=8 -march=armv4t \
			-mthumb-interwork -DSHLIndex=$(SHL_INDEX) -DCPU_TYPE=3
	ASFLAGS += -mthumb-interwork -defsym SHLIndex=$(SHL_INDEX)
	RESOURCES += $(RESOURCES_SHLHACK)
endif
endif

SIM_CC = gcc
SIM_AS = as
SIM_CFLAGS += -Wno-multichar -fshort-enums -mno-cygwin -mwindows \
		-D"__PALMOS__=1" \
		-D"_SUPPORTS_NAMESPACE=0" -D"_SUPPORTS_RTTI=1" \
		-DTARGET_PLATFORM=TARGET_PLATFORM_PALMSIM_WIN32 \
		-DTARGET_HOST=TARGET_HOST_WIN32


# Optimization settings
ifeq ($(OPTIMIZE_LEVEL), INLINING)
	PACC_OPTIMIZER_FLAG = -O5
	GCC_OPTIMIZER_FLAG = -O3
	SIM_OPTIMIZER_FLAG = -O3
else
ifeq ($(OPTIMIZE_LEVEL), INTERPROCEDURAL)
	PACC_OPTIMIZER_FLAG = -O4
	GCC_OPTIMIZER_FLAG = -O3
	SIM_OPTIMIZER_FLAG = -O3
else
ifeq ($(OPTIMIZE_LEVEL), FULL)
	PACC_OPTIMIZER_FLAG = -O3
	GCC_OPTIMIZER_FLAG = -O3
	SIM_OPTIMIZER_FLAG = -O3
else
ifeq ($(OPTIMIZE_LEVEL), INTRAPROCEDURAL)
	PACC_OPTIMIZER_FLAG = -O2
	GCC_OPTIMIZER_FLAG = -O3
	SIM_OPTIMIZER_FLAG = -O3
else
ifeq ($(OPTIMIZE_LEVEL), SOME)
	PACC_OPTIMIZER_FLAG = -O1
	GCC_OPTIMIZER_FLAG = -O1
	SIM_OPTIMIZER_FLAG = -O1
else
ifeq ($(OPTIMIZE_LEVEL), NONE)
	PACC_OPTIMIZER_FLAG = -O0
	GCC_OPTIMIZER_FLAG = -O0
	SIM_OPTIMIZER_FLAG = -O0
else
ifeq ($(OPTIMIZE_LEVEL), SIZE)
	PACC_OPTIMIZER_FLAG = -O0
	GCC_OPTIMIZER_FLAG = -Os
	SIM_OPTIMIZER_FLAG = -Os
else
# default to FULL
	PACC_OPTIMIZER_FLAG = -O3
	GCC_OPTIMIZER_FLAG = -O3
	SIM_OPTIMIZER_FLAG = -O3
endif
endif
endif
endif
endif
endif
endif

ifeq ($(TARGET_PLATFORM), device_psi)
	CFLAGS += $(PACC_OPTIMIZER_FLAG)
else
	CFLAGS += $(GCC_OPTIMIZER_FLAG)
endif
SIM_CFLAGS += $(SIM_OPTIMIZER_FLAG)

# Warning level
ifeq ($(WARNING_LEVEL), ALL)
	CFLAGS += -W -Wall
	SIM_CFLAGS += -W -Wall
else
ifeq ($(WARNING_LEVEL), NONE)
	CFLAGS += -w
	SIM_CFLAGS += -w
endif
endif

# Linker settings (must come after setting DEBUG_FLAG)
ifeq ($(TARGET_PLATFORM), device_psi)
	LD = palink
	LDFLAGS += -libpath $(SDK_LOCATION)/Libs -locals -entry _start -first _start \
			libDAL.a libBoot.a libUI.a pa_support_pure4t.a
	LINKER_STUB_SRC = $(SDK_LOCATION)/Libs/Src/LibLinkerStubPACCSHL.s
	GENLIBSTUB_MODE = pacc_shl
else
	LD = arm-palmos-gcc
	OBJCOPY = arm-palmos-objcopy
	LDFLAGS = -L $(SDK_LOCATION)/Libs -lDAL -lBoot -lUI -nostartfiles
	LINKER_STUB_SRC = $(SDK_LOCATION)/Libs/Src/LibLinkerStubGCCSHL.s
	GENLIBSTUB_MODE = gcc_shl
endif

SIM_LD = gcc
SIM_LDFLAGS += $(DEBUG_FLAG) -mno-cygwin -mwindows $(SIM_OBJ_DIR)/gcc_link.def \
		-shared -nostdlib -L $(SDK_LOCATION)/Libs -lSimDAL -lSimBoot -lSimUI -nostartfiles

LINKER_OUTPUT := $(OBJ_DIR)/$(DATABASE_NAME).axf
SIM_LINKER_OUTPUT := $(SIM_OBJ_DIR)/$(DATABASE_NAME).dll


TARGET := $(OBJ_DIR)/$(PRC_NAME)
SIM_TARGET := $(SIM_OBJ_DIR)/$(PRC_NAME)

# postlinker settings
GENLIBSTUB = $(SDK_LOCATION)/Tools/genlibstub
GENHEADER = $(SDK_LOCATION)/Tools/genheader
SHLPOSTLINK = $(SDK_LOCATION)/Tools/shlpostlink
FIXUPAMDD = $(SDK_LOCATION)/Tools/fixupamdd

###############################################################################
# Resource flags (PalmRC)
RFLAGS +=

# Default text encoding is Latin
ifeq ($(TEXTENCODING), )
	TEXTENCODING = LATIN
endif

ifeq ($(TEXTENCODING), LATIN)
	RFLAGS += -target 4.0
else
ifeq ($(TEXTENCODING), JAPANESE)
	RFLAGS += -target 4.0J
else
ifeq ($(TEXTENCODING), SIMPLIFIED_CHINESE)
	RFLAGS += -target 4.0CS
endif
endif
endif

ifeq ($(LOCALE_CHECKING), NOCHECK)
	RFLAGS += -noLocaleCheck
else
ifeq ($(LOCALE_CHECKING), STRICT)
	RFLAGS += -strictLocale
endif
endif

ifeq ($(STRICT_ENCODING), TRUE)
	RFLAGS += -strictTextEncoding
endif

###############################################################################

# function for converting sources to object file names in one of output directories
define SOURCE_LIST_TO_OBJS
	$(addprefix $(OBJ_DIR)/, $(addsuffix .o, $(foreach file, $(SOURCES), \
	$(basename $(notdir $(file))))))
endef

define LINKER_STUB_SRC_TO_OBJ
	$(addprefix $(OBJ_DIR)/, $(addsuffix .o, $(foreach file, $(LINKER_STUB_SRC), \
	$(basename $(notdir $(file))))))
endef

# function for converting sources to object file names in $(SIM_OBJ_DIR)
define SOURCE_LIST_TO_SIM_OBJS
	$(addprefix $(SIM_OBJ_DIR)/, $(addsuffix .o, $(foreach file, $(SOURCES), \
	$(basename $(notdir $(file))))))
endef

# function for converting resources into resource object files
define RESOURCE_LIST_TO_OBJS
	$(addprefix $(RSC_OBJ_DIR)/, $(addsuffix .trc, $(foreach file, $(RESOURCES), \
	$(basename $(notdir $(file))))))
endef

# function for converting stub libaries into stub archive files
define STUB_LIBRARIES_LIST_TO_ARCHIVES
	$(addprefix $(OBJ_DIR)/lib, $(addsuffix .a, $(foreach file, $(STUB_LIBRARIES), \
	$(basename $(notdir $(file))))))
endef
define STUB_LIBRARIES_LIST_TO_SIM_ARCHIVES
	$(addprefix $(SIM_OBJ_DIR)/lib, $(addsuffix .a, $(foreach file, $(STUB_LIBRARIES), \
	$(basename $(notdir $(file))))))
endef

OBJS = $(SOURCE_LIST_TO_OBJS)
SIM_OBJS = $(SOURCE_LIST_TO_SIM_OBJS)
SOURCE_PATHS += $(sort $(foreach file, $(SOURCES), $(dir $(file))))
RESOURCE_OBJS = $(RESOURCE_LIST_TO_OBJS)
RESOURCE_PATHS += $(sort $(foreach file, $(RESOURCES), $(dir $(file))))
SLD_BASENAME := $(addsuffix _Startup, $(basename $(notdir $(SLD_FILE))))
SLD_OBJ := $(addprefix $(OBJ_DIR)/, $(addsuffix .o, $(SLD_BASENAME)))
SLD_SIM_OBJ := $(addprefix $(SIM_OBJ_DIR)/, $(addsuffix .o, $(SLD_BASENAME)))
STUB_ARCHIVES = $(STUB_LIBRARIES_LIST_TO_ARCHIVES)
SIM_STUB_ARCHIVES = $(STUB_LIBRARIES_LIST_TO_SIM_ARCHIVES)
LINKER_STUB_OBJ = $(LINKER_STUB_SRC_TO_OBJ)

VPATH :=
VPATH += $(addprefix :, $(subst  ,:, $(filter-out $($(subst, :, ,$(VPATH))), $(SOURCE_PATHS) $(RESOURCE_PATHS) )))

SDK_INCLUDES = -I "$(SDK_LOCATION)/Incs" \
		-I "$(SDK_LOCATION)/PalmSourceSDK/include" \
		-I "$(SDK_LOCATION)/PalmSourceSDK/include/Core" \
		-I "$(SDK_LOCATION)/PalmSourceSDK/include/Core/Hardware" \
		-I "$(SDK_LOCATION)/PalmSourceSDK/include/Core/System" \
		-I "$(SDK_LOCATION)/PalmSourceSDK/include/Core/UI" \
		-I "$(SDK_LOCATION)/PalmSourceSDK/include/Dynamic" \
		-I "$(SDK_LOCATION)/PalmSourceSDK/include/Libraries/" \
		-I "$(SDK_LOCATION)/PalmSourceSDK/include/Extensions/" \
		-I "$(SDK_LOCATION)/PalmSourceSDK/include/Extensions/Bluetooth" \
		-I "$(SDK_LOCATION)/PalmSourceSDK/include/Extensions/ExpansionMgr" \
		-I "$(SDK_LOCATION)/GarminSDK/Incs" \
		-I "$(SDK_LOCATION)/palmOneSDK/Incs" \
		-I "$(SDK_LOCATION)/SonySDK/Incs" \
		-I "$(SDK_LOCATION)/TapwaveSDK/Incs"

LOCAL_INCLUDES = $(foreach path, $(LOCAL_INCLUDE_PATHS) $(SOURCE_PATHS) , $(addprefix -I, $(path)))
SYSTEM_INCLUDES += -I-
	SYSTEM_INCLUDES += $(foreach path, $(SYSTEM_INCLUDE_PATHS), $(addprefix -I, $(path)))
INCLUDES = $(LOCAL_INCLUDES) $(SYSTEM_INCLUDES) $(SDK_INCLUDES)

LOCAL_SIM_INCLUDES = $(foreach path, $(LOCAL_SIM_INCLUDE_PATHS) $(SOURCE_PATHS) , $(addprefix -I, $(path)))
SYSTEM_SIM_INCLUDES += -I-
SYSTEM_SIM_INCLUDES += $(foreach path, $(SYSTEM_SIM_INCLUDE_PATHS), $(addprefix -I, $(path)))
SIM_INCLUDES = $(LOCAL_SIM_INCLUDES) $(SYSTEM_SIM_INCLUDES) $(SDK_INCLUDES)

# Now add additional settings specified by user
LDFLAGS += $(ADDITIONAL_LINK_LIBRARIES)
#LDFLAGS += $(foreach option, $(ADDITIONAL_LINKER_FLAGS), $(addprefix -Wl$(comma), $(option)))
LDFLAGS += $(ADDITIONAL_LINKER_FLAGS)
CFLAGS += $(foreach define, $(DEFINES), $(addprefix -D, $(define)))
CFLAGS += $(ADDITIONAL_COMPILER_FLAGS)

SIM_LDFLAGS += $(ADDITIONAL_SIM_LINK_LIBRARIES)
SIM_LDFLAGS += $(ADDITIONAL_SIM_LINKER_FLAGS)
SIM_CFLAGS += $(foreach define, $(SIM_DEFINES), $(addprefix -D, $(define)))
SIM_CFLAGS += $(ADDITIONAL_SIM_COMPILER_FLAGS)


###############################################################################
# Flags for PRC creation

PRCFLAGS += -hostVS -dbFlagExecutable


ifeq ($(DATABASE_RESET), TRUE)
	PRCFLAGS += -dbFlagReset
endif

ifeq ($(DATABASE_BACKUP), TRUE)
	PRCFLAGS += -dbFlagBackup
endif

ifeq ($(DATABASE_HIDDEN), TRUE)
	PRCFLAGS += -dbFlagHidden
endif

ifeq ($(DATABASE_PROTECT), TRUE)
	PRCFLAGS += -dbFlagProtect
endif

ifeq ($(DATABASE_VERSION), )
	DATABASE_VERSION = 1
endif

PRCFLAGS += -dbVersion $(DATABASE_VERSION)


###############################################################################


all: $(TARGET_PLATFORM)

device_psi: $(OBJ_DIR) $(RSC_OBJ_DIR) $(TARGET)

device_gnu: $(OBJ_DIR) $(RSC_OBJ_DIR) $(TARGET)

simulator: $(SIM_OBJ_DIR) $(RSC_OBJ_DIR) $(SIM_TARGET)


# rule to create the object file directories if needed
$(OBJ_DIR):
	@[ -d $(OBJ_DIR) ] || mkdir $(OBJ_DIR) > /dev/null 2>&1

$(SIM_OBJ_DIR):
	@[ -d $(SIM_OBJ_DIR) ] || mkdir $(SIM_OBJ_DIR) > /dev/null 2>&1
	
$(RSC_OBJ_DIR):
ifneq ($(RESOURCES), )
	@[ -d $(RSC_OBJ_DIR) ] || mkdir $(RSC_OBJ_DIR) > /dev/null 2>&1
endif

# main C/C++ sources
$(OBJ_DIR)/%.o : %.c
	$(CC) -c $< $(INCLUDES) $(CFLAGS) -o $@
$(OBJ_DIR)/%.o : %.cpp
	$(CC) -c $< $(INCLUDES) $(CFLAGS) -o $@
$(OBJ_DIR)/%.o : %.cp
	$(CC) -c $< $(INCLUDES) $(CFLAGS) -o $@
$(OBJ_DIR)/%.o : %.cc
	$(CC) -c $< $(INCLUDES) $(CFLAGS) -o $@
$(OBJ_DIR)/%.o : %.C
	$(CC) -c $< $(INCLUDES) $(CFLAGS) -o $@
$(OBJ_DIR)/%.o : %.CC
	$(CC) -c $< $(INCLUDES) $(CFLAGS) -o $@
$(OBJ_DIR)/%.o : %.CPP
	$(CC) -c $< $(INCLUDES) $(CFLAGS) -o $@
$(OBJ_DIR)/%.o : %.s
	@$(AS_STATUS)
	$(AS) $< $(ASFLAGS) -o $@

# Simulator C/C++ sources
$(SIM_OBJ_DIR)/%.o : %.c
	$(SIM_CC) -c $< $(SIM_INCLUDES) $(SIM_CFLAGS) -o $@
$(SIM_OBJ_DIR)/%.o : %.cpp
	$(SIM_CC) -c $< $(SIM_INCLUDES) $(SIM_CFLAGS) -o $@
$(SIM_OBJ_DIR)/%.o : %.cp
	$(SIM_CC) -c $< $(SIM_INCLUDES) $(SIM_CFLAGS) -o $@
$(SIM_OBJ_DIR)/%.o : %.cc
	$(SIM_CC) -c $< $(SIM_INCLUDES) $(SIM_CFLAGS) -o $@
$(SIM_OBJ_DIR)/%.o : %.C
	$(SIM_CC) -c $< $(SIM_INCLUDES) $(SIM_CFLAGS) -o $@
$(SIM_OBJ_DIR)/%.o : %.CC
	$(SIM_CC) -c $< $(SIM_INCLUDES) $(SIM_CFLAGS) -o $@
$(SIM_OBJ_DIR)/%.o : %.CPP
	$(SIM_CC) -c $< $(SIM_INCLUDES) $(SIM_CFLAGS) -o $@
$(SIM_OBJ_DIR)/%.o : %.s
	@$(AS_STATUS)
	$(SIM_AS) $< $(INCLUDES) $(SIM_ASFLAGS) -o $@

$(OBJ_DIR)/lib%.a :
	"$(GENLIBSTUB)" -n $* -s $(SDK_LOCATION)/Incs/EntryNums/$*EntryNums.h \
			-m $(GENLIBSTUB_MODE) -d $(OBJ_DIR) -o $@ -i $(SHL_INDEX) $($*)

$(SIM_OBJ_DIR)/lib%.a :
	"$(GENLIBSTUB)" -n $* -s $(SDK_LOCATION)/Incs/EntryNums/$*EntryNums.h \
			-m gcc_sim -d $(SIM_OBJ_DIR) -o $@ -i $(SHL_INDEX) $($*)

$(LINKER_STUB_OBJ) : $(LINKER_STUB_SRC)
	$(AS) $(ASFLAGS) $< -o $@

# Definition file source processing
$(SLD_OBJ) :
	@echo "...Processing definition file for Device..."
	"$(GENHEADER)" -s "$(SLD_FILE)" -o $@ -t $(TYPE) -c $(CREATOR_ID) -d $(OBJ_DIR) -i $(SHL_INDEX)

$(SLD_SIM_OBJ) : 
	@echo "...Processing definition file for Simulator..."
	"$(GENHEADER)" -s "$(SLD_FILE)" -o $@ -t $(TYPE) -c $(CREATOR_ID) -d $(SIM_OBJ_DIR) -i $(SHL_INDEX) \
			-l $(SIM_OBJ_DIR)/gcc_link.def -e $(DATABASE_NAME)

# XRD source processing
$(RSC_OBJ_DIR)/%.trc : %.xrd
	"$(TOOLS_DIR)PalmRC" -quiet -i $(RESOURCE_PATHS) -p $(TARGET_FORMAT) -makeDeps $(RSC_OBJ_DIR)/$(*F).deps $(RFLAGS) -locale $(LOCALE) "`cygpath -w -a ./$<`" -o $@
$(RSC_OBJ_DIR)/%.trc : %.XRD
	"$(TOOLS_DIR)PalmRC" -quiet -i $(RESOURCE_PATHS) -p $(TARGET_FORMAT) -makeDeps $(RSC_OBJ_DIR)/$(*F).deps $(RFLAGS) -locale $(LOCALE) "`cygpath -w -a ./$<`" -o $@


# Linking step
$(LINKER_OUTPUT) : $(SLD_OBJ) $(OBJS) $(STUB_ARCHIVES) $(LINKER_STUB_OBJ)
	@echo "...Linking for Device..."
	$(LD) -o "$(LINKER_OUTPUT)" $(SLD_OBJ) $(OBJS) $(STUB_ARCHIVES) $(LINKER_STUB_OBJ) $(LDFLAGS)

$(SIM_LINKER_OUTPUT) : $(SLD_SIM_OBJ) $(SIM_OBJS) $(SIM_STUB_ARCHIVES)
	@echo "...Linking for Simulator..."
	$(SIM_LD) --entry 0 -o "$(SIM_LINKER_OUTPUT)" $(SLD_SIM_OBJ) $(SIM_OBJS) $(SIM_STUB_ARCHIVES) \
			$(SIM_LDFLAGS) $(SIM_LIBS)

ifeq ($(TARGET_PLATFORM), device_psi)
# Final target creation
$(TARGET): $(SLD_OBJ) $(LINKER_OUTPUT) $(RESOURCE_OBJS) 
	@echo "...Creating PRC for Device: $(TARGET)"
	"$(TOOLS_DIR)pelf2bin" -format hercules -directory $(OBJ_DIR) \
			-code amdc -data _mdd -rsrc 0x0000 $(LINKER_OUTPUT)
	"$(FIXUPAMDD)" -i $(OBJ_DIR)/_mdd0000.bin -o $(OBJ_DIR)/amdd0000.bin
	"$(TOOLS_DIR)elfdump" $(LINKER_OUTPUT) | "$(SHLPOSTLINK)" -c $(OBJ_DIR)/amdc0000.bin -i $(SHL_INDEX)
	"$(TOOLS_DIR)PRCMerge" -quiet -dbType $(TYPE) -dbCreator $(CREATOR_ID) \
			-dbName "$(DATABASE_NAME)" $(PRCFLAGS) \
			-o $(TARGET) $(RESOURCE_OBJS) $(OBJ_DIR)/amdd0000.bin \
			$(OBJ_DIR)/amdc0000.bin $(OBJ_DIR)/amdi0000.bin
	@echo ...Done with Device build...
	@echo 
else
$(TARGET): $(SLD_OBJ) $(LINKER_OUTPUT) $(RESOURCE_OBJS) 
	@echo "...Creating PRC for Device: $(TARGET)"
	$(OBJCOPY) -j .text -O binary $(LINKER_OUTPUT) $(OBJ_DIR)/amdc0000.bin
	"$(TOOLS_DIR)PRCMerge" -quiet -dbType $(TYPE) -dbCreator $(CREATOR_ID) \
			-dbName "$(DATABASE_NAME)" $(PRCFLAGS) \
			-o $(TARGET) $(RESOURCE_OBJS) $(OBJ_DIR)/amdc0000.bin
	@echo ...Done with Device build...
	@echo 
endif

$(SIM_TARGET): $(SLD_SIM_OBJ) $(SIM_LINKER_OUTPUT) $(RESOURCE_OBJS)
	@echo "...Creating PRC for Simulator: $(TARGET)"
	"$(TOOLS_DIR)PRCMerge" -quiet -dbType $(TYPE) -dbCreator $(CREATOR_ID) \
			-dbName "$(DATABASE_NAME)" $(PRCFLAGS) -o $(SIM_TARGET) \
			$(RESOURCE_OBJS) $(SIM_OBJ_DIR)/amdc0000.bin
	@echo ...Done with Simulator build...
	@echo

FORCE:

# Eclipse requires a clean command
# just delete everything in OBJ_DIR and SIM_OBJ_DIR and RSC_OBJ_DIR
clean :: FORCE
	-rm -rf $(OBJ_DIR)/* $(SIM_OBJ_DIR)/* $(RSC_OBJ_DIR)/*
